#define PRGNAME "AGInfo"
#define PRGVERSION "1.41"
#define PRGDATE "20030710"

/* #define DEBUG for (some) debugging output */
/* #define LARGE for Amiga, DJGPP etc. */

/*
 *  aginfo.c
 *
 *  - detects AGI v2/v3 version & game version
 *
 *  Author: Paul David Doherty <agi@pdd.de>
 *
 *  v0.9:    7 Jul 1999
 *  v1.0:    18 Jul 1999  not much changed
 *  v1.01:   17 Aug 1999  MH1/Amiga suppt; hidden list/wanted options
 *  v1.02:   24 Nov 1999  more list options
 *  v1.03:   25 Nov 1999  more possible xxDIR names (MGDIR, KQ3DIR, PQ1DIR)
 *  v1.10:    8 May 2001  new version number for millennium shift :-)
 *  v1.11:   26 Nov 2001  small typecast fix (thx to Miron Schmidt)
 *  v1.12:   28 Mar 2002  fixed bug in bestmatch algorithm
 *  v1.13:   31 Mar 2002  prepared for source code release
 *  v1.20:   26 Sep 2002  added VOL checksum handling
 *  v1.21:   27 Sep 2002  new fan-fiction detection in database
 *  v1.22:   29 Sep 2002  malloc/free memory leak fixed
 *  v1.30:   17 Oct 2002  some cleanups
 *  v1.31:    1 Nov 2002  added check for badly ordered WORDS.TOK, based
 *                        on code by Nick Sonneveld
 *  v1.32:   18 Nov 2002  some cleanups
 *  v1.33:   25 Nov 2002  small game detection fix
 *  v1.34:    2 Dec 2002  small fix
 *  v1.40:   10 Dec 2002  improved interpreter detection; file chunk
 *                        handling refactored
 *  v1.40a:   1 Jan 2003  year updated :-)
 *  v1.41:   10 Jul 2003  fclose() fixed by Claudio Matsuoka
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#ifdef __TURBOC__
#include <io.h>
#endif

/* typedefs needed by agidata.h */

typedef unsigned char type8;
typedef unsigned short type16;
typedef unsigned long int type32;

#ifdef LARGE
typedef type32 type16_32;
#define FILECHUNK 200000L
#else
typedef type16 type16_32;
#define FILECHUNK 60000L
#endif

#include "agidata.h"
#include "agicrc.h"

/* other global stuff */
#define TRUE 1
#define FALSE 0

FILE *fdi;
type8 *infiledata;

#define INTCHK 0
#define DIRCHK 1
#define VOLCHK 2

type16 int_crc = 0;
type16 dir_crc = 0;
type16 vol_crc = 0;
type32 dir_chk_df = 0L;
type16 interp_id = 0;

#define FREEDATA 0
#define KEEPDATA 1

char vol_prefix[5];

/* Amiga version string */
char *amiverstag = "\0$VER: " PRGNAME " " PRGVERSION " (" PRGDATE ")";

/*>>>>>>>>>>>>>>>>>>>>> various routines <<<<<<<<<<<<<<<<<<<<< */

/* leave the program due to error */
void
ex (char *error)
{
  fprintf (stderr, PRGNAME ": %s\n", error);
  exit (1);
}

/* called with too few/too many parameters */
void
usage (void)
{
  fprintf (stderr, PRGNAME);
  fprintf (stderr, " version " PRGVERSION);
#ifdef LARGE
  fprintf (stderr, "L");
#endif
#ifdef DEBUG
  fprintf (stderr, " [DEBUG]");
#endif
  fprintf (stderr, " (" PRGDATE ") ");

  data_usage ();

  fprintf (stderr,
	   "Copyleft 1999,2001-2003 by Paul David Doherty <agi@pdd.de>\n\n");
  fprintf (stderr, "Usage: " PRGNAME
	   "      :  detect game/interpreter in current directory\n");
  fprintf (stderr, "       " PRGNAME
	   " -ld  :  list all detected datafiles\n");
  fprintf (stderr, "       " PRGNAME
	   " -ld2 :  list detected AGIv2 datafiles\n");
  fprintf (stderr, "       " PRGNAME
	   " -ld3 :  list detected AGIv3 datafiles\n");
  fprintf (stderr, "       " PRGNAME
	   " -lf  :  list detected fan-fiction datafiles\n");
  fprintf (stderr, "       " PRGNAME
	   " -md  :  list (known) missing datafiles\n");
  exit (1);
}

void
info (void)
{
  fprintf (stdout, PRGNAME);
  fprintf (stdout, " version " PRGVERSION);
#ifdef LARGE
  fprintf (stdout, "L");
#endif
#ifdef DEBUG
  fprintf (stdout, " [DEBUG]");
#endif
  fprintf (stdout, " (" PRGDATE ") ");

  data_info ();
}

type32
do_crc (type8 crc_type, type8 data_state)
{
  type32 i;
  type16_32 bytes_read;
  type8 end_reached = 0;

  /* oh well! small memory means loading files in chunks */
  if ((infiledata = (type8 *) malloc ((type16_32) (FILECHUNK + 2L))) == NULL)
    ex ("not enough memory");

  while (end_reached == 0)
    {
      bytes_read = fread (infiledata, 1, ((type16_32) FILECHUNK), fdi);
      if (bytes_read != FILECHUNK)
	end_reached = 1;

      switch (crc_type)
	{
	case INTCHK:
	  int_crc = crc16 (infiledata, (type32) bytes_read, int_crc);
	  break;
	case DIRCHK:
	  dir_crc = crc16 (infiledata, (type32) bytes_read, dir_crc);
	  for (i = 0L; i < (type32) bytes_read; i++)
	    dir_chk_df += (type32) infiledata[i];
	  break;
	case VOLCHK:
	  vol_crc = crc16 (infiledata, (type32) bytes_read, vol_crc);
	  break;
	default:
	  break;
	}
    }

  switch (data_state)
    {
    case FREEDATA:
      /* clean up memory */
      free (infiledata);
      bytes_read = (type16_32) 0L;
      break;
    case KEEPDATA:
      /* reload first chunk; calling function must do cleanup */
      fseek (fdi, 0L, SEEK_SET);
      bytes_read = fread (infiledata, 1, ((type16_32) FILECHUNK), fdi);
      break;
    default:
      break;
    }
  return ((type32) bytes_read);
}

void
do_volstuff (void)
{
  type8 i;
  char volname[15];
  char volname_full[15];
  char numstring[5];

  strcpy (volname, vol_prefix);
  strcat (volname, "vol.");

  /* highest VOL number so far: 15 */
  /*
   * apparently, range of possible VOL numbers is:
   *  AGIv2 : 0...3 (2 bits)
   *  AGIv3 : 0...15 (4 bits)
   */
  for (i = 0; i < 20; i++)
    {
      sprintf (numstring, "%d", i);
      strcpy (volname_full, volname);
      strcat (volname_full, numstring);

      if ((fdi = fopen (volname_full, "rb")) != NULL)
	{
	  do_crc (VOLCHK, FREEDATA);
	  fclose (fdi);
	}
    }
}

type8
print_int_version (type8 platform, type32 bytes_read)
{
  type16 count = 0;
  type8 version = 1;
  type32 i, j;
  type8 int_detected = 0;

  while (version != 0)
    {
      version = int_table[count].crcvalue;
      if ((int_crc == int_table[count].crcvalue)
	  && (platform == int_table[count].platform))
	break;
      count++;
    }

  if (version != 0)
    {
      fprintf (stdout, "%s", int_table[count].interp);
      interp_id = int_table[count].int_id;
    }
  else
    {
      fprintf (stdout, "unknown ");
      int_detected = 1;

      /* find version string */
      for (i = 0L; i < (bytes_read - 10L); i++)
	{
	  if ((infiledata[i] == 'V') || (infiledata[i] == 'v'))
	    {
	      if ((infiledata[i + 1L] == 'e') && (infiledata[i + 2L] == 'r')
		  && (infiledata[i + 6L] == 'n'))
		{
		  fprintf (stdout, "(maybe: ");
		  j = i + 8L;
		  while (infiledata[j] != 0)
		    {
		      fprintf (stdout, "%c", infiledata[j]);
		      j++;
		    }
		  fprintf (stdout, ") ");
		}
	    }
	}

      /* provide CRC */
      fprintf (stdout, "[CRC:%04x]", int_crc);
    }
  fprintf (stdout, "\n");
  free (infiledata);
  return (int_detected);
}

type8
get_int_version (void)
{
  type8 i = 0;

  while (int_files[i].platform != 0)
    {
      if ((fdi = fopen (int_files[i].filename, "rb")) != NULL)
	break;
      else
	i++;
    }

  if (int_files[i].platform == 0)
    {
      /* no interpreter */
      fprintf (stdout, "AGI interpreter: none\n");
      return (2);
    }

  /* interpreter present */
  fprintf (stdout, "%s AGI interpreter: ", int_files[i].platformname);

  /* return codes: 0 = known int; 1 = unknown int; 2 = no int */
  return (print_int_version (int_files[i].platform,
			     do_crc (INTCHK, KEEPDATA)));
}

type8
check_words_tok (type32 infilelength)
{
  type32 i;
  char wordbuf[300];
  char prevword[300];
  type8 lposition, misplacedwords = 0, brokenwt = 0;

  i = 52L;
  wordbuf[0] = 0;
  prevword[0] = 0;

  while (i < (infilelength - 1L))
    {
      lposition = infiledata[i];
      i++;

      /* kludge for so-called "Tandy Demo" WORDS.TOK */
      if ((lposition != 0) && (prevword[0] == 0))
	brokenwt = 1;

      while ((infiledata[i] & 0x80) == 0)
	{
	  wordbuf[lposition] = (infiledata[i] ^ 0x7f);
	  lposition++;
	  i++;
	}
      wordbuf[lposition] = ((infiledata[i] ^ 0x7f) & 0x7f);
      lposition++;
      i++;

      wordbuf[lposition] = 0;
#ifdef DEBUG
      fprintf (stdout, "%s", wordbuf);
#endif
      if (strcmp (prevword, wordbuf) > 0)
	{
	  misplacedwords++;
#ifdef DEBUG
	  fprintf (stdout, " **bad string order***");
#endif
	}
#ifdef DEBUG
      fprintf (stdout, "\n");
#endif
      strcpy (prevword, wordbuf);
      i = i + 2;
    }

  free (infiledata);
  if (brokenwt == 0)
    return (misplacedwords);
  else
    return (0xff);
}


/*>>>>>>>>>>>>>>>>>>>>> main <<<<<<<<<<<<<<<<<<<<< */

int
main (int argc, char **argv)
{
  type16 count, bestmatchcount;
  type8 version;
  type8 unkn_int = 0, unkn_gam = 0, unkn_com = 0, unkn_vol = 0;
  type8 words_outofseq = 0;
  type8 match = 0, bestmatch = 0;

  if (argc != 1)
    {
      if (!strcmp (argv[1], "-ld2"))
	data_list (2);
      else if (!strcmp (argv[1], "-ld3"))
	data_list (3);
      else if (!strcmp (argv[1], "-ld"))
	data_list (1);
      else if (!strcmp (argv[1], "-lf"))
	data_list (11);
      else if (!strcmp (argv[1], "-bd"))
	data_list (12);
      else if (!strcmp (argv[1], "-md"))
	data_list (13);
      else
	usage ();
      return (0);
    }

  /* fix I/O streams */
#ifdef __TURBOC__
  _fmode = O_BINARY;
#endif

  /* quick game present check */
  if ((fdi = fopen ("words.tok", "rb")) == NULL)
    {
      /* user == stupid */
      fprintf (stderr, PRGNAME ": WORDS.TOK not found; no AGI game\n\n");
      usage ();
    }
  else
    fclose (fdi);

  /* okay, there is a game: print info line */
  info ();

  /* AGI interpreter version check */
  unkn_int = get_int_version ();

  /* game version check */
  strcpy (vol_prefix, "");

  if ((fdi = fopen ("words.tok", "rb")) == NULL)
    ex ("WORDS.TOK not found; no AGI game");

  words_outofseq = check_words_tok (do_crc (DIRCHK, KEEPDATA));

  if ((fdi = fopen ("object", "rb")) == NULL)
    ex ("OBJECT not found; no AGI game");
  do_crc (DIRCHK, FREEDATA);

  if ((fdi = fopen ("viewdir", "rb")) != NULL)
    {
      /* AGIv2 game */
      do_crc (DIRCHK, FREEDATA);

      if ((fdi = fopen ("logdir", "rb")) == NULL)
	ex ("LOGDIR not found; no AGI game");
      do_crc (DIRCHK, FREEDATA);

      if ((fdi = fopen ("picdir", "rb")) == NULL)
	ex ("PICDIR not found; no AGI game");
      do_crc (DIRCHK, FREEDATA);

      if ((fdi = fopen ("snddir", "rb")) == NULL)
	ex ("SNDDIR not found; no AGI game");
      do_crc (DIRCHK, FREEDATA);
    }
  else
    {
      /* presumably AGIv3 game */
      type8 dircount = 0, found = 0;

      while (dir_table[dircount].status != 0)
	{
	  if ((fdi = fopen (dir_table[dircount].maindir, "rb")) != NULL)
	    {
	      do_crc (DIRCHK, FREEDATA);
	      strcpy (vol_prefix, dir_table[dircount].prefix);
	      found = 1;
	    }
	  dircount++;
	}
      if (found == 0)
	ex ("DIR not found; no AGI game");
    }

  /* VOL checksum calculation */
  do_volstuff ();

#ifdef DEBUG
  fprintf (stdout, "[CRC:%04x/%05lx/%04x]\n", dir_crc, dir_chk_df, vol_crc);
#endif

  /* datafile version lookup with CRC & interpreter match */
  count = 0;
  version = 1;

  while (version != 0)
    {
      match = 0;
      version = version_table[count].agi_version;

      if (dir_chk_df == version_table[count].df_check)
	match++;
      if ((interp_id != 0)
	  && (interp_id == version_table[count].interp_match))
	match++;
      if (dir_crc == version_table[count].crcvalue)
	match++;
      if ((vol_crc != 0) && (vol_crc == version_table[count].volcrcvalue))
	match++;

      if (((dir_chk_df == 0L)
	   || (dir_chk_df != version_table[count].df_check))
	  && ((dir_crc == 0)
	      || (dir_crc != version_table[count].crcvalue))
	  && ((vol_crc == 0)
	      || (vol_crc != version_table[count].volcrcvalue)))
	match = 0;

      if (match > bestmatch)
	{
	  bestmatch = match;
	  bestmatchcount = count;
	}
      count++;
    }

  /* lookup finished */
  fprintf (stdout, "AGI game: ");

  if (bestmatch != 0)
    {
      count = bestmatchcount;
      version = version_table[count].agi_version;

      if ((dir_chk_df != version_table[count].df_check)
	  || (dir_crc != version_table[count].crcvalue))
	unkn_gam = 1;
      if ((interp_id != 0) && (version_table[count].interp_match != 0)
	  && (interp_id != version_table[count].interp_match))
	unkn_com = 1;
      if ((vol_crc != version_table[count].volcrcvalue)
	  && (version_table[count].volcrcvalue != 0))
	unkn_vol = 1;

      if (unkn_gam == 1)
	fprintf (stdout, "unknown; could be ");
      fprintf (stdout, "\"%s\"\n", version_table[count].game);

      if (unkn_gam == 1)
	fprintf (stdout, "[CRC:%04x/%05lx/%04x]\n",
		 dir_crc, dir_chk_df, vol_crc);
      else if ((unkn_vol == 1) && (vol_crc != 0))
	fprintf (stdout, "[VOL crc wrong: should be %04x, is %04x]\n",
		 version_table[count].volcrcvalue, vol_crc);
    }
  else
    {
      fprintf (stdout, "unknown ");
      fprintf (stdout, "[CRC:%04x/%05lx/%04x]\n",
	       dir_crc, dir_chk_df, vol_crc);
      unkn_gam = 1;
    }

  fprintf (stdout, "\nResult:\n");
  if (((unkn_int == 0) || (unkn_int == 2))
      && (unkn_gam == 0) && (unkn_com == 0) && (unkn_vol == 0))
    {
      fprintf (stdout, "- known AGIv%d game%s\n",
	       version_table[count].agi_version,
	       (unkn_int == 0) ? " and interpreter" : "");

      if (version_table[count].status == 1)
	fprintf (stdout, "- fan-made game (not an official Sierra game)\n");
      if (version_table[count].status == 2)
	fprintf (stdout, "- corrupted or hacked game version\n");
    }

  if (unkn_gam == 1)
    fprintf (stdout, "- unknown game\n");
  if (unkn_int == 1)
    fprintf (stdout, "- unknown interpreter\n");
  if (unkn_com == 1)
    fprintf (stdout, "- unknown interpreter/game combination\n");

  if (vol_crc == 0)
    fprintf (stdout, "- no VOL files present\n");
  else if (unkn_vol == 1)
    fprintf (stdout, "- unknown VOL data\n");

  if (words_outofseq > 0)
    {
      /* hardcoded kludge for padded LSL1/Macintosh */
      if ((dir_crc == 0x3147) && (vol_crc == 0xc765))
	fprintf (stdout, "- WORDS.TOK padded with garbage\n");
      else if (words_outofseq == 0xff)
	fprintf (stdout, "- non-standard WORDS.TOK file\n");
      else
	fprintf (stdout,
		 "- wrong order in WORDS.TOK due to AGI Studio bug\n");
    }

  if ((unkn_int == 1) || (unkn_gam != 0) || (unkn_com != 0)
      || (unkn_vol != 0))
    fprintf (stdout,
	     "\n"
	     "Please help me improve this program by forwarding the above\n"
	     "information to <agi@pdd.de>\n");

  return (0);
}
